/* 
 * $Author:$
 * 
 * $Id:$
 * 
 */

/*--------------------------------------------------------------------------
 * Acces routines for the Philips SJA1000 CAN device.
 *--------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>

#include "SJA1000ll.h"
#include "cansja1000.h"

/*
 * Local function can_open()
 * 
 * Parameters in *params:
 * ----------------------
 * 
 * options[] = "BRP=64;TSEG1=10;TSEG2=..."
 * 
 * BRP		Baud Rate Prescaler
 * TSEG1	Time Segment 1 
 * TSEG2	Time Segment 2
 * SAM		Sampling No
 * SJW		Sync Jump Width  
 * FMODE	1 x 32 bit or 2 x 16 bit filtering
 * FILT		Acceptance Filter Code
 * MASK		Filter Mask
 * 
 * See datasheet for possible values!
 * 
 */

static int
can_open(char *options)
{
	/* Default values: */
	static unsigned char cdr    = 0xC0;			/* PeliCAN Mode, no output, only RX0 active */
	static unsigned char mode   = 0x08;			/* default single acceptance filter			*/
	static unsigned char baud0  = 0x00;			/* default SJW=0, BRP=1						*/
	static unsigned char baud1  = 0x25;			/* default SAM=0, TSEG2=2, TSEG1=5			*/
	static unsigned long accode = 0x00000000;	/* default filter address					*/
	static unsigned long acmask = 0xFFFFFFFF;	/* default filter mask						*/
	
	/* Local vars: */
	unsigned long	val, tmp;
	char			*pTmp, sval[20];

	if (options != NULL)
	{
		pTmp = strtok(options, ";");
		while (pTmp != NULL)
		{
			if (sscanf(pTmp, "BRP=%ld", &val))
			{
				if (val < 1 || val > 64)
				{
					PRINTF(("%-15s: invalid BRP value\r\n", __FUNCTION__));
					return 4;
				}
				
				baud0 &= ~0x3F;
				baud0 |= (val - 1);
			}
			else if (sscanf(pTmp, "TSEG1=%ld", &val))
			{
				if (val < 1 || val > 16)
				{
					PRINTF(("%-15s: invalid TSEG1 value\r\n", __FUNCTION__));
					return 4;
				}
				
				baud1 &= ~0x0F;
				baud1 |= (val - 1);
			}
			else if (sscanf(pTmp, "TSEG2=%ld", &val))
			{
				if (val < 1 || val > 7)
				{
					PRINTF(("%-15s: invalid TSEG2 value\r\n", __FUNCTION__));
					return 4;
				}
				
				baud1 &= ~0x70;
				baud1 |= (val - 1) << 4;
			}
			else if (sscanf(pTmp, "SAM=%ld", &val))
			{
				if (val != 1 && val != 3)
				{
					PRINTF(("%-15s: invalid SAM value\r\n", __FUNCTION__));
					return 4;
				}
                
				baud1 &= ~0x80;
				
				if (val == 3)
					baud1 |= 0x80;
			}
			else if (sscanf(pTmp, "SJW=%ld", &val))
			{
				if (val < 1 || val > 3)
				{
					PRINTF(("%-15s: invalid SJW value\r\n", __FUNCTION__));
					return 4;
				}
				
				baud0 &= ~0xC0;
				baud0 |= (val - 1) << 6;
			}
			else if (sscanf(pTmp, "FMODE=%s", sval))
			{
				if (!strcmp(sval, "SINGLE")) mode |= 0x80;
				if (!strcmp(sval, "DUAL"))   mode &= ~0x80;
			}
			else if (sscanf(pTmp, "FILT=%lX", &val))
			{
				accode = val << 3;
			}
			else if (sscanf(pTmp, "MASK=%lX", &val))
			{
				acmask = val << 3;
			}
			else
			{	/* invalid parameter keyword */
				PRINTF(("%-15s: invalid parameter keyword!\r\n", __FUNCTION__));
				return 6;
			}
            
			pTmp = strtok(NULL, ";");
		}
	}

	can_ll_write(MODE, CANRESET);
	can_ll_write(CLKDIV, cdr);
	can_ll_write(BUSTIME0, baud0);
	can_ll_write(BUSTIME1, baud1);
	can_ll_write(ACCODE, accode >> 24 & 0xFF);
	can_ll_write(ACCODE + 1, accode >> 16 & 0xFF);
	can_ll_write(ACCODE + 2, accode >> 8 & 0xFF);
	can_ll_write(ACCODE + 3, accode & 0xFF);
	can_ll_write(ACMASK, acmask >> 24 & 0xFF);
	can_ll_write(ACMASK + 1, acmask >> 16 & 0xFF);
	can_ll_write(ACMASK + 2, acmask >> 8 & 0xFF);
	can_ll_write(ACMASK + 3, acmask & 0xFF);
	can_ll_write(OCR, CANTX0PP);

#ifdef NDEBUG
	PRINTF(("\r\n--[CAN SJA100]------------------------------------\r\n"));
	PRINTF(("   CLKDIV   = %02X\r\n", can_ll_read(CLKDIV)));
	PRINTF(("   BUSTIME0 = %02X\r\n", can_ll_read(BUSTIME0)));
	PRINTF(("   BUSTIME1 = %02X\r\n", can_ll_read(BUSTIME1)));
	PRINTF(("   STATUS   = %02X\r\n", can_ll_read(STATUS)));
	PRINTF(("   OCR      = %02X\r\n", can_ll_read(OCR)));
	PRINTF(("\r\n--------------------------------------------------\r\n"));
#endif

	can_ll_write(MODE, 0x08);				/* Switch on bus				*/
	
	trash = can_ll_read(INTRPT);			/* Clear pending interrupts		*/
	
	can_ll_write(INTEN, 0xFF);				/* Enable interrupts			*/

	can_ll_eint();						/* enable Processor's INT signal	*/
	return 0;							/* return open successfull			*/
}

/* 
 * Local function can_close()
 */
 
static void
can_close(void)
{
	/* discard your CAN-message queues or what ever you want to do... */
	can_ll_dint();						/* disable Processor's INT signal	*/
}

/* 
 * Local function can_put_frame()
 */

static void
can_put_frame(unsigned char *frame, unsigned long len)
{
	unsigned long ctr;

	for (ctr = 0; ctr < len; ctr++)
		can_ll_write(FRAME + ctr, *frame++);
    
	can_ll_write(1, COMMAND, CANTXREQ);
}

/*
 * Local function can_get_frame()
 * 
 * Returns the length of the CAN Frame.
 */

static int
can_get_frame(unsigned char *framebuffer)
{
	unsigned char infoByte;
	unsigned long ctr, len;

	*framebuffer++	= infoByte = can_ll_read(FRAME);
	len				= (infoByte & 0x0F) + 3;
	
	if (infoByte & 0x80)
		len += 2;
		
	for (ctr = 1; ctr < len; ctr++)
		*framebuffer++ = can_ll_read(FRAME + ctr);

	can_ll_read(COMMAND, CANRELRXB);
    return len;
}

/*
 * Local function can_int_event()
 * 
 * Call this within your ISR.
 * Returns the source of the CAN interrupt.
 * 
 * Returnvalues:
 * -------------
 * 	0	Unknown
 * 	1	CAN frame received
 * 	2	TX ready
 * 	3	Error Warning
 * 	4	Data Overrun
 * 	5	Wakeup
 * 	6	Error Passive
 * 	7	Arbitration Lost
 * 	8	Bus Error
 */

static int
can_int_event(void)
{
	unsigned int rVal = 0;
	
	if (info & 0x01)
		rVal = 1;
	
	if (info & 0x02)
		rVal = 2;

	if (info & 0xFC)
	{
		if (info & 0x04)
		{
			PRINTF(("CAN: Error Warning\r\n"));
			rVal = 3;
		}

		if (info & 0x08)
		{
			PRINTF(("CAN: Data Overrun\r\n"));
			rVal = 4;
		}

		if (info & 0x10)
		{
			PRINTF(("CAN: Wakeup\r\n"));
			rVal = 5;
		}

		if (info & 0x20)
		{
			PRINTF(("CAN: Error Passive\r\n"));
			rVal = 6;
		}

		if (info & 0x40)
		{
			PRINTF(("CAN: Arbitration lost\r\n"));
			rVal = 7;
		}

		if (info & 0x80)
		{
			PRINTF(("CAN: Bus Error\r\n"));
			rVal = 8;
		}
	}
	
	can_ll_eint();
	return rVal;
}
